#!/bin/bash 

# If the ## is present on any lines, the script is currently defanged.  You
# need to manually remove the double comments to have the script actually do
# anything but print out what it's going to do.  You can do this by 
# replacing all of the "\n## " strings with "\n" in the script.

### LICENSE ###

#   (C) Copyright 2008, Colorado School of Mines TICC.  All rights not
#       explicitly granted under the GNU GPLv3 are reserved.
#   Created by: Drew Hubl.
#   Maintained by: CSM TICC.
#   Special thanks to: Kenton Larson, Erich Hoover, Nick Bailey.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.



### PREFACE ###

# This works much easier if run as root, as there will be fewer prompts
# from sudo.  This is subject to change with Ubuntu Hardy Heron
# (8.04 LTS), as they are integrating the new PolicyKit from Novell.
#
# This script assumes that the wired connection is eth0, and that the
# wireless connection is eth1.  Check this before running it!
#
# Special thanks to rsalveti for his excellent getopts walkthrough at:
# http://rsalveti.wordpress.com/2007/04/03/bash-parsing-arguments-with-getopts/



### VERSION, CHANGELOG & TODO ###
# Current version:
# 0.7 - Added verbose mode option
# 0.6 - Added short option parsing
# 0.5 - Featureset extended and polished (help/usage/prompting)
# 0.4 - Done with skeleton
# 0.3 - Flow now depends on options/flags
# 0.2 - Program flow done
# 0.1 - Added dhcp3-server, tftpd-hpa config file replacing
# 0.0.5 - Sequence of commands
#
# 1 Find a way to do long options, ie: --help
# 2 List /images directory on TICC server before scping the image to localhost
# 3 Find a way to write zeroes to free space of client hard drives, if possible,
#     this will require sshing into all the slaves machines, could use:
#     http://clusterssh.wiki.sourceforge.net/Main+Page if we wanted to
# 4 Find a way to include the local imaging thing too (also uses ssh)
# 5 Add an option to create an image on the hosting computer
# 6 Stream the image to TICC
# 7 Find a better way to change config files instead of cp'ing, overwriting, 
#     and then appending ad nauseum



### HOW TO USE ###
# See image.sh -h for help and usage information



### INITIALIZE AND SET VARIABLES ###

# Initialize variables
GZIP="x"			# Controls whether or not to gzip on the fly
DHCPDREMOVE=1		# Controls whether or not to remove dhcp-server from
#                   # startup scripts
TFTPDREMOVE=1		# Controls whether or not to remove tftpd-hpa from startup
#                   # scripts
IMGDIR=0			# Controls where the image to be broadcast is stored on the
#                   # local machine
IMGFROMHD=0			# Controls where to read the image to be broadcast from
#                   # (local or ticc)
IMGPATH=0			# Controls where to put the image to be received from
#                   # another tablet
WRITEZEROES=0		# Controls whether or not to write zeroes to empty part of
#                   # drive
STOPSERVERS=0		# Controls whether or not dhcp3-server and tftpd-hpa are
#                   # shutdown during cleanup
CLEANDHCP="x"		# Controls whether or not to undo changes to dhcp config
#                   # file
CLEANTFTP="x"		# Controls whether or not to undo changes to tftpd config
#                   # file
VERBOSEMODE="0"		# Controls the verbosity of the stdout
# HELP is the variable that is displayed when the -h flag is set
# This should preserve the formatting, so it doesn't look like shit
HELP="image.sh [OPTIONS] [USERNAME] [IMAGE] [LCLIMGPATH] [LCLNETIF]

EXAMPLES:
sudo image.sh -gnstSeudzvh joe fall084200.img.gz /media/storage eth0
sudo image.sh -gnstSeudzvh joe local /media/storage eth0

PARAMETERS:
    [USERNAME] = TICC server username (root is advised against) for scping
        the image
    [IMAGE] = tablet image file on TICC server relative from /images
        directory, not the netboot image (working on listing image
        directory), specify local if pulling the image from a local
        tablet
    [LCLIMGPATH] = location to store image on host computer
    [LCLNETIF] = wired network interface (should be fully specified, ie:
        eth1), default is eth0

OPTIONS:
    -g = gzip on-the-fly to udp-sender, auto set if image ends in .gz
    -n = no gzip on-the-fly to udp-sender, either this or -g must be set
    -s = don't remove dhcp-server from the startup scripts
    -t = don't remove tftpd-hpa from the startup scripts
    -S = shutdown dhcp-server and tftpd during cleanup
    -e = don't cleanup (replace configuration file) dhcp-server after imaging
    -D = cleanup (replace configuration file) dhcp-server after imaging
    -u = don't cleanup (replace configuration file) tftp-server after imaging
    -T = cleanup (replace configuration file) tftp-server after imaging
    -d = image from a hard drive, it will prompt you for the password to
         the maxwell netboot kernel, autoset when $2 = \"local\"
    -z = write zeros to free space - not implemented yet
    -v = increase verbosity level
    -h = display this help and usage"

# Parse options
# See walkthrough located here:
# http://rsalveti.wordpress.com/2007/04/03/bash-parsing-arguments-with-getopts/
# for more information about getopts
# EXPLANATION of getopts (there aren't many good resources to learn it):
# - Init script variables before calling getopts, ie: GZIP, etc.
# - The header, ie ":gnstSeudzhq:" is the allowed flags/options that getopts
#       will recognize
# - $OPTION is the variable that getopts stores the currently evaluating option
#       in
# - $OPTION is set to "?" if getopts doesn't recognize a flag, and $OPTARG is
#       then set to the unrecognized flag
# - $@ is the full string of flags/options/parameters
# - ":" in the beginning of the header silences getopts errors, you will then
#       be forced to create/echo your own errors
# - ":" after a flag in the header specifies flags that need parameters after
#       them, such as -gst -q blarg, where q needs a parameter, if the header
#       looks like ":gnstq:", $OPTARG is set to the parameter when this happens
# - The -- and * cases should never happen, but we all know mistakes occur
# - while getopts ":gnstSeDuTdzhq:" OPTION   <-- archived for possible future
#       use if we want to implement options with parameters (haven't yet)
# Use printf instead of echo because echo takes the -n flag figuratively
# no matter what, and yes, printf is a BASH internal
if [ "0" == `printf "%s\n" "$1" | grep -c -` ]; then
	echo "Usage $0 -[g,n,s,t,S,e,D,u,T,d,z,v,h]"
	exit 65;
else
	while getopts ":gnstSeDuTdzvh" OPTION
	do
		case $OPTION in
			g)	GZIP="1";;
			n)
				if [ $GZIP == "1" ]; then
					echo "Conflicting GZIP options"
				else
					GZIP=0
				fi;;
			s)	DHCPREMOVE=0;;
			t)	TFTPDREMOVE=0;;
			S)	STOPSERVERS=1;;
			e)	CLEANDHCP=0;;
			D)	CLEANDHCP=1;;
			u)	CLEANTFTP=0;;
			T)	CLEANTFTP=1;;
			d)	IMGFROMHD=1;;
			z)	WRITEZEROES=1;;
			v)	VERBOSEMODE=1;;
			h)	echo "$HELP"
				exit 0;;
#			q)	echo "Option q"
#				echo "The parameter is $OPTARG";;
			-- )echo "No more options";;
			?)	echo "Error: Unknown option $OPTARG"
				echo "Usage: $0 -[g,n,s,t,S,e,D,u,T,d,z,v,h]"
				exit 65;;
			*)	# Default case
				echo "Error: What the heck did you do!?"
				echo "Self-destructing in 5 seconds..."
				exit -1;;
		esac
	done
fi
# Setup the wired connection
WIREDNET="x"
if [ `echo "$4" | grep -c eth[0-9]` ]; then
	WIREDNET="$4"
else
	while [ "0" == `echo $WIREDNET | grep -c eth[0-9]` ]
	do
		# Continually prompt for answer
		echo "What is the wired network interface ethX? eth[0-9]"
		read WIREDNET
	done
fi

if [[ 'local' = $2 ]]; then
	IMGFROMHD = 1
fi

# Check for .gz image file
if [[ $GZIP == "0" && "1" == `echo $2 | grep -c '.gz$'` ]]; then
	echo "File has .gz extension, overriding -n option"
	GZIP=1
fi

# Set the tftp server configuration path and its backup
TFTPDCONF=/etc/default/tftp-hpa
BAKTFTPDCONF="${TFTPDCONF}.${0##*/}.bak"

# Set the dhcp-server configuration path and its backup
DHCPCONF=/etc/dhcp3/dhcp.conf
BAKDHCPCONF="${DHCPCONF}.${0##*/}.bak"

pause()
{
	KEY=""
	echo -n $1
	stty -icanon
	KEY=`dd count=1 1>/dev/null 2> /dev/null` # `dd count=1 2>/dev/null`
	stty icanon
	echo ""
}

netinfo()
{
	# Find the IP, NetMask, and BCast addresses
	IP=`ifconfig "${1}" | awk '/inet addr/ { print $2 }'`
	NM=`ifconfig "${1}" | awk '/inet addr/ { print $4 }'`
	BC=`ifconfig "${1}" | awk '/inet addr/ { print $3 }'`

	# Cut out the parts before the ":"
	IP=${IP#*:}
	NM=${NM#*:}
	BC=${BC#*:}

	# Display cleaned output to user
	echo "Current settings for $1: "
	echo -n "IP   : "
	echo $IP
	echo -n "Mask : "
	echo $NM
	echo -n "BCast: "
	echo $BC
}

# Test for the root user, if they are not root, let them know it would
# be easier if it is run as root.
test "$(whoami)" != 'root' && (echo "You should run this with sudo, so it's\
 less annoying."

echo "Check that the wired network connection is unplugged from the switch."
pause "Press any key to continue."



### START IMAGING CONFIGURATION ###

# Get the tablet image from ticc to PWD
if [ $VERBOSEMODE == 1 ] ; then
	echo "Getting tablet image from TICC" 
fi
if [[ IMGFROMHD == 0 ]]; then
	echo ''
## 	sudo scp $1@ticc.mines.edu:/images/$2 "${IMGDIR}"
fi

# Get the netboot kernal from ticc to $IMGDIR
## echo "Getting tftpboot files from TICC"
## sudo scp $1@ticc.mines.edu:/images/tftpboot.tar.gz "${IMGDIR}"
## sudo tar -xvzf "${IMGDIR}"/tftpboot.tar.gz /var/lib

# Install the required packages
if [ "0" == `apt-cache search dhcp3-server | grep -c dhcp3-server` ]; then
	if [ $VERBOSEMODE == 0 ]; then
		echo "Installing required package dhcp3-server"
	fi
## 	sudo apt-get install dhcp3-server

	# Remove the dhcp3-server from startup services if not disallowed
	if [ "1" = $DHCPDREMOVE ]; then
		if [ $VERBOSEMODE == 1 ]; then
			echo "Removing dhcp3-server from startup services"
		fi
## 		sudo update-rc.d -f dhcp3-server remove
	fi
fi

if [ "0" = `apt-cache search tftpd-hpa | grep -c tftpd-hpa` ]; then
	if [ $VERBOSEMODE == 1 ]; then
		echo "Installing required package tftpd-hpa"
	fi
## 	sudo apt-get install tftpd-hpa

	# Remove tftpd-hpa from startup services if not disallowed
	if [ 1 = $TFTPDREMOVE ]; then
		if [ $VERBOSEMODE == 1]; then
			echo "Removing tftpd-hpa from startup services"
		fi
## 		sudo update-rc.d -f tftpd-hpa remove
	fi
fi

# Show the current network settings
# Current network settings:
netinfo $WIREDNET

# Restart the networking daemon
if [ $VERBOSEMODE == 1 ]; then
	echo "Restarting networking"
fi
## sudo /etc/init.d/networking restart

# At this point it should be okay to plug in the wired connection
echo "You may now plug in the wired connection to the switch."
pause "Press any key to continue."

# Let the user know that they should be able to use the wireless, just
# not the wired connection.
if [ $VERBOSEMODE == 1 ]; then
	echo 'You should still be able to use your wireless connection normally.'
	echo 'To reset the wired connection settings, simply go to'
	echo 'System > Administration > Network and check "Enable roaming mode."'
fi

# Backup and then change the DHCP settings for the server
if [ $VERBOSEMODE == 1 ]; then
	echo "Backing up dhcp3-server config file $DHCPCONF to $BAKDHCPCONF"
fi
## sudo cp $DHCPCONF $BAKDHCPCONF

# Write the config file for the dhcp server
# Not the most elegant, but it works
if [ $VERBOSEMODE == 1 ]; then
	echo ''
## 	echo "Writing config file for dhcp server"
fi
## echo 'subnet 192.168.0.0 netmask 255.255.255.0 {' 1>$DHCPCONF
## echo '   range 192.168.0.10 192.168.0.254 ;' 1>>$DHCPCONF
## echo '   option subnet-mask 255.255.255.0;' 1>>$DHCPCONF
## echo '   option routers 192.168.0.1;' 1>>$DHCPCONF
## echo '   option broadcast-address 192.168.0.255 ;' 1>>$DHCPCONF
## echo '   default-lease-time 30000;' 1>>$DHCPCONF
## echo '   max-lease-time 30000;' 1>>$DHCPCONF
## echo '   server-name "192.168.0.1";' 1>>$DHCPCONF
## echo '   filename "pxelinux.0";' 1>>$DHCPCONF
## echo '}' 1>>$DHCPCONF
## echo '' 1>>$DHCPCONF
## echo "# This file automatically generated by ${0##*/}." 1>>$DHCPCONF
## echo "# The backup for this file is saved at $BAKTFTPDCONF" 1>>$DHCPCONF

# Show the current network settings
echo "New network settings: "
netinfo "$WIREDNET"

# Restart the DHCP server
# It may already be running, so a start might not catch these changes.
# A restart will, and will simply fail at stopping the server (because
# it's not running)
if [ $VERBOSEMODE == 1 ]; then
	echo "Restarting dhcp-server..."
fi
## sudo /etc/init.d/dhcp3-server restart

# Show the current network settings
echo "Newest network settings: "
netinfo "$WIREDNET"

# Backup tftpd-hpa config file
if [ $VERBOSEMODE == 1 ]; then
	echo "Backing up tftpd config file $TFTPDCONF to $BAKTFTPDCONF"
fi
## sudo cp $TFTPDCONF $BAKTFTPDCONF

# Write the config file for the tftpd-hpa server
# Not the most elegant but it works.
if [ $VERBOSEMODE == 1 ]; then
	echo ''
## 	echo "Writing config file for tftp server"
fi
## echo '# Defaults for tftpd-hpa' 1>$TFTPDCONF
## echo 'RUN_DAEMON="yes"' 1>>$TFTPDCONF
## echo 'OPTIONS="-l -s /var/lib/tftpboot"' 1>>$TFTPDCONF
## echo '' 1>>$TFTPDCONF
## echo "# This file automatically generated by ${0##*/}." 1>>$TFTPDCONF
## echo "# The backup for this file is saved at $BAKTFTPDCONF" 1>>$TFTPDCONF

# Restart the tftp server
# It may already be running, so a start may not catch these changes.
# A restart will, and will simply fail at stopping the server (because
# it's not running.
if [ $VERBOSEMODE == 1 ]; then
	echo "Restarting tftpd-hpa server"
fi
## sudo /etc/init.d/tftpd-hpa restart"

# Instruct the tablet network boot now
echo "You may now network-boot the tablets."
pause 'Press any key to continue.'

# Check for gzip (-g) or no-gzip(-n) flag
while [[ 0 != $GZIP && 1 != $GZIP ]]
do
	# Continually prompt for answer
	echo "Do you want to gzip on-the-fly?"
	read GZIP

	case $GZIP in
		y*)
			GZIP=1
			;;
		n*)
			GZIP=0
			;;
		*)
			echo $'\n'"You have to answer this."
			;;
	esac
done

echo "Done with configuration, waiting to image."
echo "Double check all tablets are netbooted."
echo "THIS IS YOUR LAST CHANCE TO CHANGE ANY OPTIONS!"
pause 'Press any key to continue.'



### START IMAGING PROCESS ###

# Save the multicasted image to the hard drive on the host (this) computer
#if [[ 1 == $IMGTOHOST ]]; then
#	# Prompt user for imformation
#	while [[ 0 == $IMGPATH ]]
#	do
#		# Continually prompt for answer
#		echo "Where do you want to save the image? (foobar.img.gz, n to quit)"
#		read $IMGPATH
#	done
#	if [ "n" != "$IMGPATH" ]; then
#		udp-receiver --file "${IMGPATH}"
#	else
#		exit -1;
#	fi
#fi

# Enable if local install is chosen
if [[ 1 == $IMGFROMHD ]]; then
	if [ $VERBOSEMODE == 1 ]; then
		echo "You cannot terminate ssh or it will kill the imaging process."
	fi
	IMGSERVERIP=0
	while [[ 0 == $IMGSERVERIP ]]
	do
		# Continually prompt for answer
		echo "What is the IP address of the image-serving tablet?"
		read IMGSERVERIP
	done
	ssh "root@${IMGSERVERIP}"
	killall udp-receiver
	cat /dev/hda | gzip -c | udp-sender
fi

# Send the image to udp-sender
if [ $VERBOSEMODE == 1 ]; then
	echo "Sending image to udp-sender"
fi
echo "Double check the number of clients"
case $GZIP in
	"1")
		if [ "0"!="$IMGDIR" ]; then
			echo ""
## 			cat "${IMGDIR}" | gzip -c | udp-sender --interface "eth${WIREDNET}"
		fi
		;;
	"0")
## 		udp-sender --file "${IMGDIR}" --interface "${WIREDNET}"
		;;
esac



### CLEANUP ###

echo "Cleaning up"

# Prompt for replacing the dhcp-server configuration files
while [[ 1 != $CLEANDHCP && 0 != $CLEANDHCP ]]
do
	echo "Do you want to replace the original dhcp-server config files? (y/n)"
	read CLEANDHCP

	case $CLEANDHCP in
		y*)
			CLEANDHCP=1;;
		n*)
			CLEANDHCP=0;;
		*)
			echo $'\n'"You have to answer this.";;
	esac
done

if [ $CLEANDHCP = 1 ]; then
	if [ $VERBOSEMODE == 1 ]; then
		echo "Replacing original dhcp3-server config file $DHCPCONF"
	fi
## 	sudo cp $BAKDHCPCONF $DHCPCONF

else
	echo "Original dhcp3-server config file located at $BAKDHCPCONF"
fi

# Prompt for replacing the tftpd-hpa configuration files
while [[ 1 != $CLEANTFTP && 0 != $CLEANTFTP ]]
do
	echo "Do you want to replace the original tftpd-hpa config files? (y/n)"
	read CLEANTFTP

	case $CLEANTFTP in
		y*)
			CLEANTFTP=1;;
		n*)
			CLEANTFTP=0;;
		*)
			echo $'\n'"You have to answer this.";;
	esac
done

if [ $CLEANTFTP = 1 ]; then
	if [ $VERBOSEMODE == 1 ]; then
		echo "Replacing original tftpd-hpa config file 2$TFTPDCONF"
	fi
## 	sudo cp $BAKTFTPDCONF $TFTPDCONF
else
	echo "Original dhcp3-server config file located at $BAKDHCPCONF"
fi

if [[ 1 == $STOPSERVERS ]]; then
	if [ $VERBOSEMODE == 1 ]; then
		echo "Stopping dhcp3-server"
	fi
## 	sudo /etc/init.d/dhcp3-server stop
	if [ $VERBOSEMODE == 1 ]; then
		echo "Stopping tftpd-hpa"
	fi
## 	sudo /etc/init.d/tftpd-hpa stop
else
	if [ $VERBOSEMODE == 1 ]; then
		echo "Restarting dhcp3-server"
	fi
## 	sudo /etc/init.d/dhcp3-server restart
	if [ $VERBOSEMODE == 1 ]; then
		echo "Restarting tftpd-hpa"
	fi
## 	sudo /etc/init.d/tftpd-hpa restart
fi			

if [ 1="$WRITEZEROES" ]; then
	echo "To write zeroes to the free space on the client machines, simply "
	echo "cat /dev/zero > ./zero"
	echo "and delete the file later on by "
	echo "rm ./zero"
fi

echo "Done imaging."

exit 0

### Phew.  That was cool. ####
### And FWIW: Fuck Microsoft! ####

